import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal
import pint
from pint.util import UnitsContainer
import pytest

from iam_units import convert_gwp, format_mass, registry


DEFAULTS = pint.get_application_registry()

# Parameters for test_units(), tuple of:
# 1. A literal string to be parsed as a unit.
# 2. Expected dimensionality of the parsed unit.
# 3. True if the units are not in pint's default_en.txt, but are defined in
#    definitions.txt.
energy = UnitsContainer({'[energy]': 1})
PARAMS = [
    ('toe', energy, False),
    ('tce', energy, True),
    ('GW a', energy, False),
    ('kWa', energy, True),
    ('EUR_2005', UnitsContainer({'[currency]': 1}), True),
]


@pytest.mark.parametrize('unit_str, dim, new_def', PARAMS,
                         ids=lambda v: v if isinstance(v, str) else '')
def test_units(unit_str, dim, new_def):
    if new_def:
        # Units defined in dimensions.txt are not recognized by base pint
        with pytest.raises(pint.UndefinedUnitError):
            DEFAULTS('1 ' + unit_str)

    # Physical quantity of units is recognized
    dim = registry.get_dimensionality(dim)

    # Units can be parsed and have the correct dimensionality
    assert registry('1 ' + unit_str).dimensionality == dim


def test_orders_of_magnitude():
    # The registry recognizes units prefixed by an order of magnitude
    assert registry('1.2 billion EUR').to('million EUR').magnitude == 1.2e3


def test_kt():
    # The registry should correctly interpret `kt` as a weight (not velocity)
    assert str(registry('1000 kt').to('Mt')) == '1.0 megametric_ton'

    # A default UnitRegistry should interpret `kt` as velocity
    with pytest.raises(pint.DimensionalityError):
        pint.UnitRegistry()('kt').to('Mt')


def test_emissions_internal():
    # Dummy units can be created
    registry('0.5 _gwp').dimensionality == {'[_GWP]': 1.0}

    # Context can be enabled
    with registry.context('AR5GWP100', __a=1.0) as r:
        # Context-specific values are activated
        assert r('a_CH4 * a_N2O').to_base_units().magnitude == 28 * 265

        # Context-specific conversion rules are available
        r('0.5 t').to('_gwp')


@pytest.mark.parametrize('units', [
    't {}',               # Mass
    'Mt {} / a',          # Mass rate
    'kt {} / (ha * yr)',  # Mass flux
])
@pytest.mark.parametrize('metric, species_in, species_out, expected_value', [
    ('AR5GWP100', 'CH4', 'CO2', 28),
    ('AR5GWP100', 'CH4', 'CO2e', 28),
    ('AR4GWP100', 'CH4', 'CO2', 25),
    ('SARGWP100', 'CH4', 'CO2', 21),

    # Same-species conversion with metric=None and compatible names
    (None, 'CO2', 'CO2_eq', 1.),
    (None, 'CO2eq', 'CO2e', 1.),

    # Species names which are substrings of one another match correctly
    ('AR5GWP100', 'HFC143', 'CO2', 328.0),
    ('AR5GWP100', 'HFC143a', 'CO2', 4800.0),
])
def test_convert_gwp(units, metric, species_in, species_out, expected_value):
    # Bare masses can be converted
    qty = registry.Quantity(1.0, units.format(''))
    expected = registry(f'{expected_value} {units}')
    assert convert_gwp(metric, qty, species_in, species_out) == expected

    # '[mass] [speciesname] (/ [time])' can be converted; the input species is
    # extracted from the *qty* argument
    qty = f'1.0 ' + units.format(species_in)
    expected = registry(f'{expected_value} {units}')
    assert convert_gwp(metric, qty, species_out) == expected

    # Tuple of (vector magnitude, unit expression) can be converted where the
    # the unit expression contains the input species name
    arr = np.array([1.0, 2.5, 0.1])
    qty = (arr, units.format(species_in))
    expected = arr * expected_value

    # Conversion works
    result = convert_gwp(metric, qty, species_out)
    # Magnitudes are as expected
    assert_array_almost_equal(result.magnitude, expected)


def test_convert_gwp_carbon():
    # CO2 can be converted to C
    qty = (44. / 12, 'tonne CO2')
    result = convert_gwp('AR5GWP100', qty, 'C')
    assert result.units == registry('tonne')
    assert_almost_equal(result.magnitude, 1.0)

    # C can be converted to CO2
    qty = (1, 'tonne C')
    expected = registry.Quantity(44. / 12, 'tonne')
    assert convert_gwp('AR5GWP100', qty, 'CO2e') == expected


@pytest.mark.parametrize('units_in, species_str, spec, output', [
    ('Mt', 'CO2', None, 'megametric_ton CO2'),
    ('kg / year', 'CO2', None, 'kilogram CO2 / year'),
    # Compact spec: no spaces around '/'
    ('kg / year', 'CO2', ':C', 'kilogram CO2/year'),
    # Abbreviated/symbol spec AND Unicode subscript
    ('Mt', 'CO₂', ':~', 'Mt CO₂'),
    ('kg / year', 'CO₂', ':~', 'kg CO₂ / a'),
    # Abitrary string, e.g. including a hint at the metric name
    ('gram / a', 'CO₂-e (AR4)', ':~', 'g CO₂-e (AR4) / a'),
])
def test_format_mass(units_in, species_str, spec, output):
    # Quantity object can be formatted
    qty = registry.Quantity(3.5, units_in)
    assert format_mass(qty, species_str, spec) == output

    # Unit object can be formatted
    qty = registry.Unit(units_in)
    assert format_mass(qty, species_str, spec) == output



from itertools import chain
from pathlib import Path

import pandas as pd
import sys


# Base path for package code
BASE_PATH = Path(__file__).parent

# Base path for package data
DATA_PATH = BASE_PATH / 'data'


# Format strings for emissions()
_EMI_HEADER = """# This file was generated using:
#    python -m iam_units.update emissions
# DO NOT ALTER THIS FILE MANUALLY!
"""

# Format string for individual metrics. To expand the set of supported
# conversions, duplicate and modify the first pair of lines in the context.
# Only the portion before of the ':' needs to be modified. Currently supported:
# 1. Mass
# 2. Mass rate, i.e. mass per time.
# 3. Flux, i.e. mass per area per time.
_EMI_DATA = f"""{_EMI_HEADER}
@context(_a=NaN) {{metric}}
    [mass] -> [_GWP]: value * (_a * _gwp / kg)
    [_GWP] -> [mass]: value / (_a * _gwp / kg)
    [mass] / [time] -> [_GWP] / [time]: value * (_a * _gwp / kg)
    [_GWP] / [time] -> [mass] / [time]: value / (_a * _gwp / kg)
    [mass] / [time] / [area] -> [_GWP] / [time] / [area]: value * (_a * _gwp / kg)
    [_GWP] / [time] / [area] -> [mass] / [time] / [area]: value / (_a * _gwp / kg)


    {{defs}}
@end
"""  # noqa: E501

# Format string for an importable Python module defining the *pattern* regex,
# which resembles: (?<=[ -])(CO2|C|N2O|CH4)(?=[ -/]|[^\w]|$)
# - Preceded by a space or '-' character.
# - Followed by a space, '-', '/', end-of-string, or non-word (\w) character.
#   The latter avoids matching only the 'C' within 'CH4'.
_EMI_CODE = fr"""{_EMI_HEADER}
import re

# All recognised emission species usable with convert_gwp(). See *pattern*.
SPECIES = [
    '{{symbols}}',
    ]

# Sets of symbols that refer to the same species and are interchangeable.
EQUIV = [
    set({{equiv}}),
    ]

# Regular expression for one *SPECIES* in a pint-compatible unit string.
pattern = re.compile(
    '(?<=[ -])('
    + '|'.join(SPECIES)
    + r')(?=[ -/]|[^\w]|$)')
"""

# Equivalents: different symbols for the same species.
_EMI_EQUIV = [
    ['CO2', 'CO2_eq', 'CO2e', 'CO2eq'],
    ['C', 'Ce'],
]


def emissions():
    """Update emissions definitions files."""
    data_path = DATA_PATH / 'emissions'

    # - Load data.
    # - Sort by symbol.
    # - Convert to long form with 'metric' and 'value' columns.
    # - Drop missing values.
    data = pd.read_csv(data_path / 'metric_conversions.csv') \
             .sort_values('Symbol') \
             .melt(id_vars=['Species', 'Symbol'], var_name='metric') \
             .dropna(subset=['value'])

    # List of symbols requiring a GWP context to covert
    symbols = sorted(data['Symbol'].unique())

    # Format and write the species defs file
    lines = [_EMI_HEADER]
    for group in _EMI_EQUIV:
        lines.extend(f'a_{s} = a_{group[0]}' for s in group[1:])
    lines.extend(f'a_{s} = NaN' for s in symbols)
    lines.append('')
    (data_path / 'species.txt').write_text('\n'.join(lines))

    # Write a Python module with a regex matching the species names

    # Prepare list including all symbols
    all_symbols = list(chain(*_EMI_EQUIV, symbols))

    # Format and write
    code = _EMI_CODE.format(
        symbols="',\n    '".join(all_symbols),
        equiv="),\n    set(".join(map(repr, _EMI_EQUIV)),
    )
    (BASE_PATH / 'emissions.py').write_text(code)

    # Write one file containing a context for each metric
    for metric, _data in data.groupby('metric'):
        # Conversion factor definitions
        defs = [f'a_{row.Symbol} = {row.value}' for _, row in _data.iterrows()]

        # Format the template with the definitions
        content = _EMI_DATA.format(metric=metric, defs='\n    '.join(defs))

        # Write to file
        (data_path / f'{metric}.txt').write_text(content)


if __name__ == '__main__':
    # Invoked using 'python -m iam_units.update'
    # For each additional argument, call the function of the same name
    for module in sys.argv[1:]:
        locals()[module]()



# This file was generated using:
#    python -m iam_units.update emissions
# DO NOT ALTER THIS FILE MANUALLY!

import re

# All recognised emission species usable with convert_gwp(). See *pattern*.
SPECIES = [
    'CO2',
    'CO2_eq',
    'CO2e',
    'CO2eq',
    'C',
    'Ce',
    'C10F18',
    'C2F6',
    'C3F8',
    'C4F10',
    'C5F12',
    'C6F14',
    'CCl4',
    'CF4',
    'CFC11',
    'CFC113',
    'CFC114',
    'CFC115',
    'CFC12',
    'CFC13',
    'CH2Cl2',
    'CH3Br',
    'CH3CCl3',
    'CH3Cl',
    'CH4',
    'CHCl3',
    'HCFC123',
    'HCFC124',
    'HCFC141b',
    'HCFC142b',
    'HCFC21',
    'HCFC22',
    'HCFC225ca',
    'HCFC225cb',
    'HCFE235da2',
    'HFC125',
    'HFC134',
    'HFC134a',
    'HFC143',
    'HFC143a',
    'HFC152',
    'HFC152a',
    'HFC161',
    'HFC227ea',
    'HFC23',
    'HFC236cb',
    'HFC236ea',
    'HFC236fa',
    'HFC245ca',
    'HFC245fa',
    'HFC32',
    'HFC365mfc',
    'HFC41',
    'HFC4310mee',
    'HFE125',
    'HFE134',
    'HFE143a',
    'HFE227ea',
    'HFE236ca12',
    'HFE236ea2',
    'HFE236fa',
    'HFE245cb2',
    'HFE245fa1',
    'HFE245fa2',
    'HFE263fb2',
    'HFE329mcc2',
    'HFE338mcf2',
    'HFE338pcc13',
    'HFE347mcc3',
    'HFE347mcf2',
    'HFE347pcf2',
    'HFE356mec3',
    'HFE356pcc3',
    'HFE356pcf2',
    'HFE356pcf3',
    'HFE374pc2',
    'HFE4310pccc124',
    'HFE449sl',
    'HFE569sf2',
    'Halon1201',
    'Halon1211',
    'Halon1301',
    'Halon2402',
    'N2O',
    'NF3',
    'PFPMIE',
    'SF5CF3',
    'SF6',
    'cC3F6',
    'cC4F8',
    ]

# Pairs of emission species symbols that are interchangeable.
EQUIV = [
    set(['CO2', 'CO2_eq', 'CO2e', 'CO2eq']),
    set(['C', 'Ce']),
    ]

# Regular expression for one *SPECIES* in a pint-compatible unit string.
pattern = re.compile(
    '(?<=[ -])('
    + '|'.join(SPECIES)
    + r')(?=[ -/]|[^\w]|$)')
